from tic.reputation.shared.dtos.reputation_item import (
    ReputationItem,
    Reputation,
    ReputationTags,
    ReputationCategory,
    ScanDetails,
    SignatureInfo,
    Candidate,
    Mfg,
)
from ..enums.signature_status import SignatureStatus
from ....reputation.rep_reff_feed_consumer.raw_reputation_message import ReffReputation
from ....shared.apc.enums.apc_verdict import ApcVerdict
from ....shared.apc.enums.apc_category import ApcCategory
from ....shared.apc.enums.apc_status import ApcStatus
from .apc import Apc
from ....shared.utils.date_util import (
    utc_now,
    match_str_strptime_to_datetime,
)


class ReffReputationItem(ReputationItem):
    decode_settings = {
        "cb_first_seen": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "feed_date": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "categorization_date": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "date_modified": lambda x: utc_now()
        if x is None
        else match_str_strptime_to_datetime(x),
        "reputation": lambda x: Reputation()
        if x is None
        else Reputation.from_dict(x, True, True),
        "scan_details": lambda x: None
        if x is None
        else ScanDetails.from_dict(x, True, True),
        "mfg": lambda x: None if x is None else Mfg.from_dict(x, True, True),
        "signature_info": lambda x: None
        if x is None
        else SignatureInfo.from_dict(x, True, True),
        "sources": lambda x: list() if x is None else x,
        "apc": lambda x: None if x is None else Apc.from_dict(x, True, True),
        "candidate": lambda x: None
        if x is None
        else Candidate.from_dict(x, True, True),
        "reputation_v2": lambda x: Reputation()
        if x is None
        else Reputation.from_dict(x, True, True),
        "reff": lambda x: ReffReputation()
        if x is None
        else ReffReputation.from_dict(x, True, True),
    }

    def __init__(
        self,
        sha256=None,
        md5=None,
        sha1=None,
        tlsh=None,
        sources=None,
        reputation=None,
        file_name=None,
        cb_first_seen=None,
        feed_date=None,
        categorization_date=None,
        date_modified=None,
        scan_details=None,
        mfg=None,
        signature_info=None,
        apc: Apc = None,
        candidate=None,
        version=0,
        cynic=None,
        reff=None,
        reputation_v2=None,
    ):
        super().__init__(
            sha256,
            md5,
            sha1,
            tlsh,
            sources,
            reputation,
            file_name,
            cb_first_seen,
            feed_date,
            categorization_date,
            date_modified,
            scan_details,
            mfg,
            signature_info,
            apc,
            candidate,
            version,
        )
        self.cynic = cynic
        self.reff = ReffReputation() if reff is None else reff
        self.reputation_v2 = Reputation() if reputation_v2 is None else reputation_v2

    def is_definitively_categorized(self):
        return self.reputation_v2.is_locked() or self.reputation_v2.category in {
            ReputationCategory.TRUSTED_WHITE_LIST,
            ReputationCategory.KNOWN_MALWARE,
        }

    def categorize(self):
        # we can assume here that there will always be either CDCR, MFG, or APC data because of the new model--
        # should never get forwarded truly empty data

        # Safeguard here in case defaults do not work for some reason.
        if self.reputation_v2 is None:
            self.reputation_v2 = Reputation()

        # make reputation.comment a set so we can add and remove items easily
        # will cast back to list at the end of categorization
        self.reputation_v2.comment = set()

        self.reputation_v2.comment.add(ReputationTags.REFF)  # Adding REFF tag

        # begin categorizing
        self._categorize_trust_threat()

        # Add locked tag whether we need to use MFG to update category or not.
        # May not be used for category since MFG can't be locked without rep also locked.
        if self.mfg is not None and self.mfg.locked is True:
            self.reputation_v2.comment.add(ReputationTags.MFG_LOCKED)

        # check for whether we need MFG added
        if (
            self.reputation_v2.is_locked()
            or self.reputation_v2.category == ReputationCategory.TRUSTED_WHITE_LIST
        ):
            self.reputation_v2.comment.add(ReputationTags.DEFINITIVE_BEFORE_MFG)
            return
        if self.mfg is not None:
            self.reputation_v2.comment.add(ReputationTags.MFG_DATA_EVALUATED)
            self._categorize_with_mfg()

        self.reputation_v2.comment = list(self.reputation_v2.comment)
        if self.apc is not None:
            self._perform_apc_adjustments()
        self.categorization_date = utc_now()

    def _common_white_update(self):
        """
        updates category from ADAPTIVE to COMMON if no positives and first seen > 30 days or is validly signed.
        """
        if (
            self.scan_details is None
            or self.signature_info is None
            or self.cb_first_seen is None
        ):
            return
        elif self.scan_details.positives == 0 and self.scan_details.total_scans > 0:
            first_seen = (
                self.scan_details.first_scanned_date
                if self.scan_details.first_scanned_date < self.cb_first_seen
                else self.cb_first_seen
            )
            if (
                self.signature_info.signature_status
                == SignatureStatus.SIGNED_AND_VERIFIED
                or (utc_now() - first_seen).days >= 30
            ):
                self.reputation_v2.category = ReputationCategory.COMMON_WHITE_LIST
                self.reputation_v2.comment.add(ReputationTags.COMMON_WHITE_CRITERIA_MET)
        return

    def _categorize_trust_threat(self):
        if self.reputation_v2.threat == 100:
            self.reputation_v2.category = ReputationCategory.KNOWN_MALWARE
            self.reputation_v2.comment.add(ReputationTags.CB_HIGH_THREAT)
        elif self.reputation_v2.threat >= 50:
            self.reputation_v2.category = ReputationCategory.PUP
            self.reputation_v2.comment.add(ReputationTags.CB_PUP)
        else:
            self.reputation_v2.category = ReputationCategory.ADAPTIVE_WHITE_LIST
            if (
                self.scan_details is not None
                and self.scan_details.positives is not None
                and not self.reputation_v2.is_locked()
            ):
                self.reputation_v2.comment.add(ReputationTags.CB_SCORES_INCONCLUSIVE)
                if self.scan_details.positives >= 6:
                    self.reputation_v2.comment.add(
                        ReputationTags.SUSPECT_FROM_POSITIVES
                    )
                    self.reputation_v2.category = ReputationCategory.SUSPECT_MALWARE
                else:
                    self.reputation_v2.category = ReputationCategory.ADAPTIVE_WHITE_LIST
            if self.reputation_v2.category == ReputationCategory.ADAPTIVE_WHITE_LIST:
                self._common_white_update()

        # handles when only trust is locked or when threat locked low and trust high
        if self.reputation_v2.trust >= 7:
            self.reputation_v2.category = ReputationCategory.TRUSTED_WHITE_LIST
            self.reputation_v2.comment.add(ReputationTags.CB_HIGH_TRUST)

    def _categorize_with_mfg(self):
        # Relying on MFG locked only if reputation locked-
        # We should already have category based on rep lock.

        # add US_CERT tag if UsCert in mfg scanners.
        if "UsCert" in self.mfg.classification_scanners:
            self.reputation_v2.comment.add(ReputationTags.US_CERT_MALWARE)

        # use MFG PUP or malware_family of Mimikatz or LaZange to classify
        if not self.reputation_v2.is_locked():
            if (
                self._classification_scanners_count > 2
                or "Avira" in self.mfg.classification_scanners
            ):
                # Classify as malware if malware_family is Mimikatz or LaZange
                if (
                    self.mfg.malware_family.result == "Mimikatz"
                    or self.mfg.malware_family.result == "LaZange"
                ):
                    # Update the signature and malware group
                    self.mfg.malware_group.result = "PWS"
                    self.mfg.signature = (
                        "PWS:"
                        + (self.mfg.platform.result if self.mfg.platform.result else "")
                        + "/"
                        + self.mfg.malware_family.result
                        + (("." + ".".join(self.mfg.tags)) if self.mfg.tags else "")
                    )
                    if "Avira" not in self.mfg.classification_scanners:
                        self.reputation_v2.category = ReputationCategory.SUSPECT_MALWARE
                        self.reputation_v2.comment.add(ReputationTags.MFG_SUSPECT)
                        return
                    else:
                        self.reputation_v2.category = ReputationCategory.KNOWN_MALWARE
                        self.reputation_v2.comment.add(
                            ReputationTags.MFG_AVIRA_CONVICTION
                        )
                        return
                # use MFG PUP if caught by MFG
                elif self.mfg.malware_group.result in {"Adware", "PUP"}:
                    self.reputation_v2.category = ReputationCategory.PUP
                    self.reputation_v2.comment.add(ReputationTags.MFG_PUP)
                    return

        if not self.is_definitively_categorized():
            if (
                self._classification_scanners_count > 2
                or "Avira" in self.mfg.classification_scanners
            ):
                if "Avira" not in self.mfg.classification_scanners:
                    self.reputation_v2.category = ReputationCategory.SUSPECT_MALWARE
                    self.reputation_v2.comment.add(ReputationTags.MFG_SUSPECT)
                    return
                else:
                    self.reputation_v2.category = ReputationCategory.KNOWN_MALWARE
                    self.reputation_v2.comment.add(ReputationTags.MFG_AVIRA_CONVICTION)
                    return
            # no data to convict in MFG
            self.reputation_v2.comment.add(ReputationTags.MFG_NO_CONVICTION)
            # don't downgrade suspect, pup classifications from trust/threat categorization
            if self.reputation_v2.category not in {
                ReputationCategory.SUSPECT_MALWARE,
                ReputationCategory.PUP,
            }:
                self.reputation_v2.category = ReputationCategory.ADAPTIVE_WHITE_LIST
                self._common_white_update()

    def _perform_apc_adjustments(self):
        self.reputation_v2.comment = set(self.reputation_v2.comment)
        self.reputation_v2.comment.add(ReputationTags.APC_DATA_EVALUATED)

        _used_apc_data = False

        # check for already counted avira data in MFG
        if self.mfg and "Avira" in self.mfg.classification_scanners:
            self.reputation_v2.comment.add(ReputationTags.APC_AVIRA_ALREADY_IN_MFG)
            self.reputation_v2.comment = list(self.reputation_v2.comment)
            return

        # Check if more recent Avira data than APC data (if there's no query data use Avira data)
        if "avira" in self.sources and (
            not self.apc.query_date or self.feed_date > self.apc.query_date
        ):
            self.reputation_v2.comment.add(ReputationTags.APC_AVIRA_ALREADY_IN_MFG)
            self.reputation_v2.comment = list(self.reputation_v2.comment)
            return

        # check that avira data is ready
        if self.apc.status != ApcStatus.OK:
            self.reputation_v2.comment.add(ReputationTags.APC_UNREADY)
            self.reputation_v2.comment = list(self.reputation_v2.comment)
            return

        # check that avira has a result
        if self.apc.category == ApcCategory.UNKNOWN:
            self.reputation_v2.comment.add(ReputationTags.APC_UNKNOWN)
            self.reputation_v2.comment = list(self.reputation_v2.comment)
            return

        # upgrade categorization
        if (
            self.apc.verdict == ApcVerdict.PUP
            and self.reputation_v2.category != ReputationCategory.SUSPECT_MALWARE
        ):
            _used_apc_data = True
            self.reputation_v2.category = ReputationCategory.PUP
            self.reputation_v2.comment.add(ReputationTags.APC_PUP)
        elif self.apc.verdict == ApcVerdict.MALWARE:
            _used_apc_data = True
            self.reputation_v2.category = ReputationCategory.KNOWN_MALWARE
            self.reputation_v2.comment.add(ReputationTags.APC_MALICIOUS)

        if _used_apc_data:
            self.reputation_v2.comment.add(ReputationTags.APC_USED)
        else:
            if self.reputation_v2.category == ReputationCategory.NOT_LISTED:
                self.reputation_v2.category = ReputationCategory.ADAPTIVE_WHITE_LIST
            self.reputation_v2.comment.add(ReputationTags.APC_UNUSED)

        self.reputation_v2.comment = list(self.reputation_v2.comment)
